// -*- C++ -*-

/* 
 * GChemPaint
 * atom.cc
 *
 * Copyright (C) 2001-2003
 *
 * Developed by Jean Brfort <jean.brefort@ac-dijon.fr>
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include "config.h"
#include "atom.h"
#include "bond.h"
#include "globals.h"
#include "settings.h"
#include "view.h"
#include "widgetdata.h"
#include <math.h>
#include <chemistry/element.h>

using namespace gcu;

gcpAtom::gcpAtom(): Atom()
{
	m_Valence = -1; //unspecified
	m_nve = 0;
	m_nH = 0;
	m_HPos = GetBestSide();
	m_ChargeAuto = false;
}

gcpAtom::~gcpAtom()
{
}

gcpAtom::gcpAtom(int Z, double x, double y, double z): Atom(Z, x, y, z)
{
	m_ChargeAuto = false;
	m_HPos = GetBestSide();
	SetZ(Z);
}

#ifdef HAVE_OPENBABEL
gcpAtom::gcpAtom(OBAtom* atom, double zoom): Atom()
{
	m_x = atom->GetX() * zoom;
	m_y = - atom->GetY() * zoom;
	m_z = atom->GetZ() * zoom;
	SetZ(atom->GetAtomicNum());
	gchar* Id = g_strdup_printf("a%d", atom->GetIdx());
	SetId(Id);
	g_free(Id);
//	m_Valence = atom->GetValence();
	m_HPos = true;
}
#endif

void gcpAtom::SetZ(int Z)
{
	Atom::SetZ(Z);
	m_Valence = Element::GetElement(m_Z)->GetDefaultValence();
	// compute valence electrons number: FIXME: should be moved elsewhere
	if (m_Valence > 0)
	{
		if (Z < 3) m_nve = Z;
		else if (Z < 11) m_nve = Z - 2;
		else if (Z < 19) m_nve = Z - 10;
		else if (Z < 37) m_nve = Z - 18;
		else if (Z < 55) m_nve = Z - 36;
		else if (Z < 87) m_nve = Z - 54;
		else m_nve = Z - 86;
		Update();
		m_HPos = GetBestSide();
	}
	else m_nH = 0;
}

int gcpAtom::GetTotalBondsNumber()
{
	std::map<Atom*, Bond*>::iterator i;
	int n = 0;
	for (i = m_Bonds.begin(); i != m_Bonds.end(); i++)
		n += (*i).second->GetOrder();
	return n;
}

double gcpAtom::Distance(gcpAtom* pAtom)
{
	return sqrt(square(m_x - pAtom->m_x) + square(m_y - pAtom->m_y) + square(m_z - pAtom->m_z));
}

void gcpAtom::zoom(double ZoomFactor)
{
	m_x *= ZoomFactor;
	m_y *= ZoomFactor;
	m_z *= ZoomFactor;
}

void gcpAtom::AddBond(gcpBond* pBond)
{
	Atom::AddBond(pBond);
	Update();
}

void gcpAtom::RemoveBond(gcpBond* pBond)
{
	Atom::RemoveBond(pBond);
	Update();
}

bool gcpAtom::GetBestSide()
{
	if (m_Bonds.size() == 0) return Element::BestSide(m_Z);
	std::map<Atom*, Bond*>::iterator i;
	double sum = 0.0;
	for (i = m_Bonds.begin(); i != m_Bonds.end(); i++)
		sum -= cos(((gcpBond*)(*i).second)->GetAngle2DRad(this));
	if (fabs(sum) > 0.1) return (sum >= 0.0);
	else return Element::BestSide(m_Z);
}

void gcpAtom::Update()
{
	if (m_Valence <= 0) return;//FIXME:stupid
	if (m_ChargeAuto)
	{
		m_Charge = 0;
		m_ChargeAuto = false;
	}
	int nb, nlp = (m_nve - m_Valence) / 2;
	if ((m_Charge > 0) && (nlp > 0)) nlp --;
	nb = m_nve - 2 * nlp - m_Charge;
	if (nb + nlp > 4) nb -= 2; //octet rule
	m_nH = nb - GetTotalBondsNumber();
	if ((! m_Charge) && (m_nH == -1))
	{
		m_Charge = (nlp)? 1: -1;
		m_ChargeAuto = true;
	}
	if (m_nH < 0) m_nH = 0;
	m_HPos = GetBestSide();
}
	
bool gcpAtom::IsInCycle(gcpCycle* pCycle)
{
	map<Atom*, Bond*>::iterator i;
	for (i = m_Bonds.begin(); i != m_Bonds.end(); i++)
		if (((gcpBond*)(*i).second)->IsInCycle(pCycle)) return true;
	return false;
}
void gcpAtom::Add(GtkWidget* w)
{
	if (!w) return;
	gcpWidgetData* pData = (gcpWidgetData*)g_object_get_data(G_OBJECT(w), "data");
	double x, y, x1, y1;
	m_width =  m_height = 2.0 * pData->Padding;
	GetCoords(&x, &y);
	x *= pData->ZoomFactor;
	y *= pData->ZoomFactor;
	double dFontHeight = pData->View->GetFontHeight();
	char * szFontName = pData->View->GetFontName();
	GnomeCanvasItem* item;
	GnomeCanvasGroup* group;
	gint width, height;
	PangoContext* pc = pData->View->GetPangoContext();
	group = GNOME_CANVAS_GROUP(gnome_canvas_item_new(pData->Group, gnome_canvas_group_get_type(), NULL));
	g_signal_connect(G_OBJECT(group), "event", G_CALLBACK(on_event), w);
	g_object_set_data(G_OBJECT(group), "object", this);
	if ((GetZ() != 6) || (GetBondsNumber() == 0))
	{
		const gchar* symbol = GetSymbol();
		PangoLayout *pl1, *pl = pango_layout_new(pc);
		pango_layout_set_text(pl, symbol, strlen(symbol));
		PangoRectangle rect;
		pango_layout_get_extents(pl, &rect, NULL);
		m_width = rect.width / PANGO_SCALE;
		m_height =  rect.height / PANGO_SCALE;
		item = gnome_canvas_item_new(
							group,
							gnome_canvas_rect_get_type(),
							"x1", x - (double)m_width / 2 - pData->Padding,
							"y1", y - dFontHeight / 2 - pData->Padding,
							"x2", x + (double)m_width / 2 + pData->Padding,
							"y2", y + dFontHeight / 2 + pData->Padding,
							"fill_color", "white",
							NULL);
		g_signal_connect(G_OBJECT(item), "event", G_CALLBACK(on_event), w);
		g_object_set_data(G_OBJECT(item), "object", this);
		item = gnome_canvas_item_new(
							group,
							gnome_canvas_text_get_type(),
							"text", GetSymbol(),
							"x", rint(x),
							"y", rint(y),
							"font", szFontName,
							"anchor", GTK_ANCHOR_CENTER,
							"fill_color", (pData->IsSelected(this))? SelectColor: Color,
							NULL);
		g_signal_connect(G_OBJECT(item), "event", G_CALLBACK(on_event), w);
		g_object_set_data(G_OBJECT(item), "object", this);
		int n = GetAttachedHydrogens();
		if (n > 0)
		{
//			GtkAnchorType anchor;
//			double xpos;
			gchar *tmp;
			if (m_HPos)
			{
				x = rint(x + m_width / 2 + 3);
				pango_layout_set_text(pl, "H", 1);
				pango_layout_get_extents(pl, &rect, NULL);
				width = rect.width / PANGO_SCALE;
				height =  rect.height / PANGO_SCALE;
				item = gnome_canvas_item_new(
									group,
									gnome_canvas_rect_get_type(),
									"x1", x - 1,
									"y1", y - dFontHeight / 2 - pData->Padding,
									"x2", x + (double)width + pData->Padding,
									"y2", y + dFontHeight / 2 + pData->Padding,
									"fill_color", "white",
									NULL);
				g_signal_connect(G_OBJECT(item), "event", G_CALLBACK(on_event), w);
				g_object_set_data(G_OBJECT(item), "object", this);
				item = gnome_canvas_item_new(
									group,
									gnome_canvas_text_get_type(),
									"text", "H",
									"x", rint(x),
									"y", rint(y),
									"font", szFontName,
									"anchor", GTK_ANCHOR_W,
									"fill_color", (pData->IsSelected(this))? SelectColor: Color,
									NULL);
				g_signal_connect(G_OBJECT(item), "event", G_CALLBACK(on_event), w);
				g_object_set_data(G_OBJECT(item), "object", this);
				if (n > 1)
				{
					pango_context_set_font_description(pc, pData->View->GetPangoSmallFontDesc());
					pl1 = pango_layout_new(pc);
					x += width + 3;
					y1 = rint(y + dFontHeight * 0.4);
					tmp = g_strdup_printf("%d", n);
					pango_layout_set_markup(pl1, tmp, strlen(tmp));
					pango_layout_get_extents(pl1, &rect, NULL);
					width = rect.width / PANGO_SCALE;
					height =  rect.height / PANGO_SCALE;
					item = gnome_canvas_item_new(
										group,
										gnome_canvas_rect_get_type(),
										"x1", x - 1,
										"y1", y1 - dFontHeight / 3 - pData->Padding,
										"x2", x + (double)width + pData->Padding,
										"y2", y1 + dFontHeight / 3 + pData->Padding,
										"fill_color", "white",
										NULL);
					g_signal_connect(G_OBJECT(item), "event", G_CALLBACK(on_event), w);
					g_object_set_data(G_OBJECT(item), "object", this);
					item = gnome_canvas_item_new(
										group,
										gnome_canvas_text_get_type(),
										"text", tmp,
										"x", rint(x),
										"y", y1,
										"font", pData->View->GetSmallFontName(),
										"anchor", GTK_ANCHOR_W,
										"fill_color", (pData->IsSelected(this))? SelectColor: Color,
										NULL);
					g_signal_connect(G_OBJECT(item), "event", G_CALLBACK(on_event), w);
					g_object_set_data(G_OBJECT(item), "object", this);
					pango_context_set_font_description(pData->View->GetPangoContext(), pData->View->GetPangoFontDesc());
					g_free(tmp);
				}
			}
			else
			{
				x = rint(x - m_width / 2 - 1);
				if (n > 1)
				{
					pango_context_set_font_description(pc, pData->View->GetPangoSmallFontDesc());
					pl1 = pango_layout_new(pc);
					y1 = rint(y + dFontHeight * 0.4);
					tmp = g_strdup_printf("%d", n);
					pango_layout_set_markup(pl1, tmp, strlen(tmp));
					pango_layout_get_extents(pl1, &rect, NULL);
					width = rect.width / PANGO_SCALE;
					height =  rect.height / PANGO_SCALE;
					item = gnome_canvas_item_new(
										group,
										gnome_canvas_rect_get_type(),
										"x1", x - (double)width - pData->Padding,
										"y1", y1 - dFontHeight / 3 - pData->Padding,
										"x2", x + 1,
										"y2", y1 + dFontHeight / 3 + pData->Padding,
										"fill_color", "white",
										NULL);
					g_signal_connect(G_OBJECT(item), "event", G_CALLBACK(on_event), w);
					g_object_set_data(G_OBJECT(item), "object", this);
					item = gnome_canvas_item_new(
										group,
										gnome_canvas_text_get_type(),
										"text", tmp,
										"x", rint(x),
										"y", y1,
										"font", pData->View->GetSmallFontName(),
										"anchor", GTK_ANCHOR_E,
										"fill_color", (pData->IsSelected(this))? SelectColor: Color,
										NULL);
					g_signal_connect(G_OBJECT(item), "event", G_CALLBACK(on_event),w);
					g_object_set_data(G_OBJECT(item), "object", this);
					pango_context_set_font_description(pc, pData->View->GetPangoFontDesc());
					g_free(tmp);
					x -= width;
				}
				pango_layout_set_text(pl, "H", 1);
				pango_layout_get_extents(pl, &rect, NULL);
				width = rect.width / PANGO_SCALE;
				height =  rect.height / PANGO_SCALE;
				item = gnome_canvas_item_new(
									group,
									gnome_canvas_rect_get_type(),
									"x1", x - (double)width - pData->Padding,
									"y1", y - dFontHeight / 2 - pData->Padding,
									"x2", x + 1,
									"y2", y + dFontHeight / 2 + pData->Padding,
									"fill_color", "white",
									NULL);
				g_signal_connect(G_OBJECT(item), "event", G_CALLBACK(on_event), w);
				g_object_set_data(G_OBJECT(item), "object", this);
				item = gnome_canvas_item_new(
									group,
									gnome_canvas_text_get_type(),
									"text", "H",
									"x", rint(x),
									"y", rint(y),
									"font", szFontName,
									"anchor", GTK_ANCHOR_E,
									"fill_color", (pData->IsSelected(this))? SelectColor: Color,
									NULL);
				g_signal_connect(G_OBJECT(item), "event", G_CALLBACK(on_event), w);
				g_object_set_data(G_OBJECT(item), "object", this);
//				anchor = GTK_ANCHOR_E;
//				xpos = rint(x - width / 2 - 1);
			}
/*			gchar* tmp;
			if (n > 1) tmp = g_strdup_printf("H<sub>%d</sub>", n);
			else tmp = g_strdup("H");
		pango_layout_set_markup(pl, tmp, strlen(tmp));
		pango_layout_get_extents(pl, &rect, NULL);
printf("height:%d\n", rect.height / PANGO_SCALE);
			item = gnome_canvas_item_new(
								group,
								gnome_canvas_text_get_type(),
								"markup", tmp,
								"x", xpos,
								"y", rint(y + (rect.height / PANGO_SCALE - height) / 2),//- rect.height / PANGO_SCALE),
								"font", szFontName,
								"anchor", anchor,
								"fill_color", (pData->IsSelected(this))? SelectColor: Color,
								NULL);
			g_free(tmp);*/
		}
	}
	else
	{
		m_nH = 0;//if symbol is not drawn, its attached hydrogens will not be either
		item = gnome_canvas_item_new(
								group,
								gnome_canvas_rect_get_type(),
								"x1", x - 3,
								"y1", y - 3,
								"x2", x + 3,
								"y2", y + 3,
								"fill_color", "white",
								NULL);
		gnome_canvas_request_redraw((GnomeCanvas*)w, (int)x-3, (int)y-3, (int)x+3, (int)y+3);
		gnome_canvas_item_lower_to_bottom(GNOME_CANVAS_ITEM(group));
		gnome_canvas_item_raise(GNOME_CANVAS_ITEM(group), 1);
		g_signal_connect(G_OBJECT(item), "event", G_CALLBACK(on_event), w);
		g_object_set_data(G_OBJECT(item), "object", this);
	}
	pData->Items[this] = group;
	m_width /= pData->ZoomFactor;
	m_height /= pData->ZoomFactor;
	char charge = GetCharge();
	if (charge)
	{
		int align = GetChargePosition(x, y);
		x *= pData->ZoomFactor;
		y *= pData->ZoomFactor;
		switch (align)
		{
			case -2:
				y += 6.0;//FIXME: change 6.0 end other constants to variables
				break;
			case -1:
				x -= 6.0;
				break;
			case 1:
				x += 6.0;
				break;
			case 2:
				y -= 6.0;
				break;
		}
		GnomeCanvasGroup* chargegroup = GNOME_CANVAS_GROUP(gnome_canvas_item_new(group, gnome_canvas_group_get_type(), NULL));
		g_object_set_data ((GObject*)group, "charge", chargegroup);
		gnome_canvas_item_new(
						chargegroup,
						gnome_canvas_ellipse_get_type(),
						"width_units", 1.0,
						"outline_color", (pData->IsSelected(this))? SelectColor: Color,
						"x1", x - 4.0,
						"x2", x + 4.0,
						"y1", y - 4.0,
						"y2", y + 4.0,
						NULL);
		GnomeCanvasPoints *points = gnome_canvas_points_new (2);
		points->coords[0] = x - 2.5;
		points->coords[1] = points->coords[3] = y;
		points->coords[2] = x + 2.5;
		gnome_canvas_item_new(
						chargegroup,
						gnome_canvas_line_get_type(),
						"points", points,
						"fill_color", (pData->IsSelected(this))? SelectColor: Color,
						"width_units", 1.0,
						NULL);
		if (charge > 0)
		{	
			points->coords[0] = points->coords[2] = x;
			points->coords[1] = y - 2.5;
			points->coords[3] = y + 2.5;
			gnome_canvas_item_new(
							chargegroup,
							gnome_canvas_line_get_type(),
							"points", points,
							"fill_color", (pData->IsSelected(this))? SelectColor: Color,
							"width_units", 1.0,
							NULL);
		}
	}
//	gnome_canvas_item_get_bounds(GNOME_CANVAS_ITEM(group), &x, &y, &x1, &y1);
//	gnome_canvas_request_redraw((GnomeCanvas*)w, (int)x, (int)y, (int)x1, (int)y1);
}//FIXME: finish this function

void gcpAtom::Update(GtkWidget* w)
{
	if (!w) return;
	gcpWidgetData* pData = (gcpWidgetData*)g_object_get_data(G_OBJECT(w), "data");
}//FIXME: write this function

#define CHARGE_NE 1
#define CHARGE_NW 2
#define CHARGE_N 4
#define CHARGE_SE 8
#define CHARGE_SW 16
#define CHARGE_S 32
#define CHARGE_E 64
#define CHARGE_W 128

int gcpAtom::GetChargePosition(double& x, double& y)
{
	list<double> AngleList;
	list<double>::iterator n;
	double angle;
	unsigned char def;
	if (((GetZ() != 6 || m_Bonds.size() == 0)) && m_nH)
	{
		if (m_HPos)
		{
			def = 0xB6;
			AngleList.push_front(225.0);
			AngleList.push_front(135.0);
		}
		else
		{
			def = 0x6D;
			AngleList.push_front(315.0);
			AngleList.push_front(45.0);
		}
	}
	else def = 0xff;
	map<Atom*, Bond*>::iterator i = m_Bonds.begin();
	while (i != m_Bonds.end())
	{
		n = AngleList.begin();
		angle = ((gcpBond*)(*i).second)->GetAngle2D(this) + 180.0;
		while ((n != AngleList.end()) && (*n < angle)) n++;
		AngleList.insert(n, angle);
		i++;
		if ((def & CHARGE_NE) && (angle >= 180.0) && (angle <= 270.0)) def -= CHARGE_NE;
		if ((def & CHARGE_NW) && (((angle >= 270.0) && (angle <= 360.0)) || (fabs(angle) < 0.1))) def -= CHARGE_NW;
		if ((def & CHARGE_N) && (angle >= 225.0) && (angle <= 315.0)) def -= CHARGE_N;
		if ((def & CHARGE_SE) && (angle >= 90.0) && (angle <= 180.0)) def -= CHARGE_SE;
		if ((def & CHARGE_SW) && (((angle >= 0.0) && (angle <= 90.0)) || (fabs(angle - 360.0) < 0.1))) def -= CHARGE_SW;
		if ((def & CHARGE_S) && (angle >= 45.0) && (angle <= 135.0)) def -= CHARGE_S;
		if ((def & CHARGE_E) && ((angle <= 225.0) && (angle >= 135.0))) def -= CHARGE_E;
		if ((def & CHARGE_W) && (angle >= 315.0) || (angle <= 45.0)) def -= CHARGE_W;
	}
	if (def)
	{
		if (def & CHARGE_NE)
		{
			x = m_x + m_width / 2.0;
			y = m_y - m_height / 2.0;
			return 1;
		}
		if (def & CHARGE_NW)
		{
			x = m_x - m_width / 2.0;
			y = m_y - m_height / 2.0;
			return -1;
		}
		if (def & CHARGE_N)
		{
			x = m_x;
			y = m_y - m_height / 2.0;
			return 2;
		}
		if (def & CHARGE_SE)
		{
			x = m_x + m_width / 2.0;
			y = m_y + m_height / 2.0;
			return 1;
		}
		if (def & CHARGE_SW)
		{
			x = m_x - m_width / 2.0;
			y = m_y + m_height / 2.0;
			return -1;
		}
		if (def & CHARGE_S)
		{
			x = m_x;
			y = m_y + m_height / 2.0;
			return -2;
		}
		if (def & CHARGE_E)
		{
			x = m_x + m_width / 2.0;
			y = m_y;
			return 1;
		}
		if (def & CHARGE_W)
		{
			x = m_x - m_width / 2.0;
			y = m_y;
			return -1;
		}
	}
	AngleList.push_back((angle = AngleList.front()) + 360.0);
	double dir = 0.0, max = 0.0;
	//if we are there, there is at least two bonds
	for (n = AngleList.begin(), n++; n != AngleList.end(); n++)
	{
		if (*n - angle > max)
		{
			if (*n - angle - max > 0.1) x = (*n + angle) / 2;
			if (m_nH)
			{
				if (m_HPos && ((x > 225.0) || (x < 135.0))) dir = x;
				else if (m_HPos && (x > 45.0) && (x < 315.0)) dir = x;
			}
			else dir = x;
			max = *n - angle;
		}
		angle = *n;
	}
	max = sqrt(square(m_width) + square(m_height)) / 2.0 + 24;//Could do beter, should replace 24 by something more intelligent
	x = m_x + max * cos(dir / 180.0 * M_PI);
	y = m_y - max * sin(dir / 180.0 * M_PI);
	return 0;
}

bool gcpAtom::LoadNode(xmlNodePtr)
{
	SetZ(GetZ());
	return true;
}

